//! Indexes for the subscription manager
//!
//! This module contains the definition of the indexes used by the subscription
//! manager to provide a convenient way to choose indexes to reduce the number
//! of checks before sending the events to the subscribers.
use nostr_rs_types::types::{filter::TagValue, Event, Filter, Id, Kind};
use std::collections::HashSet;

#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
/// Events can be indexed by different fields.
///
/// This index is a multi-value index, meaning that it can be indexed by multiple fields
pub enum Index {
    /// Author ID
    Author(Id),
    /// Note ID
    Id(Id),
    /// Note Kind
    Kind(Kind),
    /// Tag
    Tag(String, TagValue),
    /// A catch-all index
    Anything,
}

impl Index {
    /// Converts an event into a list of indexes
    pub fn from(event: &Event) -> Vec<Self> {
        let mut subscriptions = vec![];

        subscriptions.push(Index::Author(event.author().to_owned()));
        subscriptions.push(Index::Id(event.id.to_owned()));

        for t in event.tags() {
            if let Some(v) = t.get_indexable_value() {
                subscriptions.push(Index::Tag(t.get_identifier().to_owned(), v));
            }
        }

        subscriptions.push(Index::Kind(event.kind()));
        subscriptions.push(Index::Anything);
        subscriptions
    }
}

/// Indexes for the storage engine.
#[derive(Debug, Clone)]
pub enum CompoundIndex {
    /// Index by tag.
    Tag(String, HashSet<TagValue>),
    /// Index by id.
    Id(Vec<Id>),
    /// Index by author.
    Author(Vec<Id>),
    /// Index by kind.
    Kind(Vec<Kind>),
    /// Table scan.
    TableScan,
}

impl CompoundIndex {
    /// Splits the index into a list of single indexes.
    pub fn split(self) -> Vec<Index> {
        match self {
            CompoundIndex::Tag(tag, tags) => tags
                .into_iter()
                .map(|tag_value| Index::Tag(tag.clone(), tag_value))
                .collect(),
            CompoundIndex::Id(ids) => ids.into_iter().map(Index::Id).collect(),
            CompoundIndex::Author(authors) => authors.into_iter().map(Index::Author).collect(),
            CompoundIndex::Kind(kinds) => kinds.into_iter().map(Index::Kind).collect(),
            CompoundIndex::TableScan => vec![Index::Anything],
        }
    }
}

impl From<&mut Filter> for CompoundIndex {
    fn from(query: &mut Filter) -> Self {
        if !query.ids.is_empty() {
            CompoundIndex::Id(std::mem::take(&mut query.ids))
        } else if !query.authors.is_empty() {
            CompoundIndex::Author(std::mem::take(&mut query.authors))
        } else if !query.tags.is_empty() {
            // find the tag with fewer options to iterate over
            // and convert that tag as the index.
            // This block of code is slightly different, because only an index is needed, instead of a list of indexes
            // to iterate over, therefore the index with fewer options is selected.
            let mut tag_with_fewer_opts = query
                .tags
                .iter()
                .map(|(tag, values)| (tag.to_owned(), values.len()))
                .collect::<Vec<_>>();

            tag_with_fewer_opts.sort_by_key(|(_, len)| *len);
            let (key, _) = tag_with_fewer_opts.remove(0);

            if let Some(tags) = query.tags.remove(&key) {
                CompoundIndex::Tag(key, tags)
            } else {
                CompoundIndex::TableScan
            }
        } else if !query.kinds.is_empty() {
            CompoundIndex::Kind(std::mem::take(&mut query.kinds))
        } else {
            CompoundIndex::TableScan
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use serde_json::json;

    #[test]
    fn test_index_preference_ids_over_others() {
        let mut filter: Filter = serde_json::from_value(json!({
            "ids": ["42224859763652914db53052103f0b744df79dfc4efef7e950fc0802fc3df3c5"],
            "authors": ["42224859763652914db53052103f0b744df79dfc4efef7e950fc0802fc3df3c5"],
            "#r": ["foo"],
            "kinds": [1]
        }))
        .expect("filter");

        let index = CompoundIndex::from(&mut filter);
        assert!(matches!(index, CompoundIndex::Id(_)));
    }

    #[test]
    fn test_index_preference_authors_over_tags_and_kinds() {
        let mut filter: Filter = serde_json::from_value(json!({
            "authors": ["42224859763652914db53052103f0b744df79dfc4efef7e950fc0802fc3df3c5"],
            "#r": ["foo"],
            "kinds": [1]
        }))
        .expect("filter");

        let index = CompoundIndex::from(&mut filter);
        assert!(matches!(index, CompoundIndex::Author(_)));
    }

    #[test]
    fn test_index_preference_tags_over_kinds() {
        let mut filter: Filter = serde_json::from_value(json!({
            "#r": ["foo"],
            "kinds": [1]
        }))
        .expect("filter");

        let index = CompoundIndex::from(&mut filter);
        assert!(matches!(index, CompoundIndex::Tag(_, _)));
    }

    #[test]
    fn test_index_preference_kinds_when_others_are_empty() {
        let mut filter: Filter = serde_json::from_value(json!({
            "kinds": [1]
        }))
        .expect("filter");

        let index = CompoundIndex::from(&mut filter);
        assert!(matches!(index, CompoundIndex::Kind(_)));
    }
}
