//! # Nostr Protocol Types. Request
//!
//! Used to request events and subscribe to new updates.
use crate::{common::SerializeDeserialize, types};
use hex::ToHex;
use once_cell::sync::Lazy;
use rand::RngCore;
use serde_json::Value;
use std::collections::VecDeque;

static ALL_EVENTS_PREFIX: Lazy<String> = Lazy::new(|| {
    let mut prefix = [0u8; 4];
    rand::thread_rng().fill_bytes(&mut prefix);
    prefix.encode_hex::<String>()
});

/// Request: used to request events and subscribe to new updates.
///
/// More details at https://github.com/nostr-protocol/nips/blob/master/01.md#communication-between-clients-and-relays
#[derive(Debug, Clone)]
pub struct Subscribe {
    /// The new subscription ID. If not provided, a default one will be created
    /// automatically
    pub subscription_id: types::SubscriptionId,
    /// List of filters to requests events and subscribe to new updates
    /// Events that match any of the filters are to be returned, i.e., multiple filters are to be interpreted as || conditions.
    pub filters: Vec<types::Filter>,
}

/// Checks if the subscription ID is for all events
pub fn is_all_events(subscription_id: &types::SubscriptionId) -> bool {
    subscription_id.starts_with(&*ALL_EVENTS_PREFIX)
}

impl Subscribe {
    /// Creates a new subscription with a random ID to subscribe to all events
    pub fn to_all_events() -> Self {
        Self {
            subscription_id: types::SubscriptionId::with_prefix(&ALL_EVENTS_PREFIX)
                .expect("valid subscription id"),
            filters: vec![Default::default()],
        }
    }
}

impl From<types::Filter> for Subscribe {
    fn from(filter: types::Filter) -> Self {
        Self {
            filters: vec![filter],
            ..Default::default()
        }
    }
}

impl From<Vec<types::Filter>> for Subscribe {
    fn from(filters: Vec<types::Filter>) -> Self {
        Self {
            filters,
            ..Default::default()
        }
    }
}

impl Default for Subscribe {
    fn default() -> Self {
        Self {
            subscription_id: types::SubscriptionId::default(),
            filters: vec![types::Filter::default()],
        }
    }
}

impl SerializeDeserialize for Subscribe {
    fn get_tag() -> &'static str {
        "REQ"
    }

    fn serialize(&self) -> Result<Vec<Value>, String> {
        let mut r = vec![
            Value::String(Self::get_tag().to_owned()),
            Value::String((*self.subscription_id).to_owned()),
        ];

        for element in self.filters.iter() {
            r.push(serde_json::to_value(element).map_err(|e| e.to_string())?);
        }

        Ok(r)
    }

    fn deserialize(mut array: VecDeque<Value>) -> Result<Self, String>
    where
        Self: Sized,
    {
        let subscription_id = array
            .pop_front()
            .ok_or("Missing element 1 in the array")?
            .as_str()
            .ok_or("Invalid type for element 1, expecting a string")?
            .try_into()
            .map_err(|e: types::subscription_id::Error| {
                format!("Invalid subscription id: {}", e)
            })?;

        Ok(Subscribe {
            subscription_id,
            filters: if !array.is_empty() {
                serde_json::from_value::<Vec<types::Filter>>(serde_json::Value::Array(
                    array.into_iter().collect::<Vec<serde_json::Value>>(),
                ))
                .map_err(|e: serde_json::Error| e.to_string())?
            } else {
                vec![]
            },
        })
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::Request as ClientRequest;
    use chrono::Utc;

    #[test]
    fn encoding_decoding() {
        let r: ClientRequest = Subscribe {
            filters: vec![
                types::Filter {
                    authors: vec![
                        "npub1h0h2zfcp3z7acvc9l0qvpgfqznkalj77whsdkmmlwanfssys3ngsx336a8"
                            .try_into()
                            .expect("some addr object"),
                    ],
                    until: Some(Utc::now()),
                    limit: 30,
                    ..Default::default()
                },
                types::Filter {
                    limit: 20,
                    ..Default::default()
                },
            ],
            ..Default::default()
        }
        .into();

        let serialized = serde_json::to_string(&r).expect("valid json string");
        let obj: ClientRequest = serde_json::from_str(&serialized).expect("valid request");
        let obj = obj.as_request().expect("request");
        let r = r.as_request().expect("request");
        assert_eq!(r.subscription_id, obj.subscription_id);
        assert_eq!(r.filters.len(), obj.filters.len());
    }

    #[test]
    fn test_subscribe_all() {
        let x = Subscribe::to_all_events();
        assert!(is_all_events(&x.subscription_id));
        assert!(!is_all_events(&Default::default()));
    }
}
