//! OK
//!
//! When submitting events to relays, clients currently have no way to know if
//! an event was successfully committed to the database. This NIP introduces the
//! concept of command results which are like NOTICE's except provide more
//! information about if an event was accepted or rejected.
use crate::{
    common::SerializeDeserialize,
    types::{self, Addr},
};
use serde_json::Value;
use std::collections::VecDeque;

#[derive(Clone, PartialEq, Eq, Debug)]
/// ROkStatus
pub enum ROkStatus {
    /// Everything is ok
    Ok,
    /// Proof of work failed
    Pow,
    /// Event is known already
    Duplicate,
    /// User is blocked
    Blocked,
    /// User is blacklisted
    Blacklisted,
    /// Rate limit exceeded
    RateLimit,
    /// Invalid event
    Invalid,
    /// Error
    Error(String),
}

impl From<(Addr, ROkStatus)> for ROk {
    fn from((id, status): (Addr, ROkStatus)) -> Self {
        Self { id, status }
    }
}

impl ROkStatus {
    /// Compute readable status string
    pub fn computer_readable(&self) -> String {
        match self {
            ROkStatus::Ok => "",
            ROkStatus::Pow => "pow",
            ROkStatus::Duplicate => "duplicate",
            ROkStatus::Blocked => "blocked",
            ROkStatus::Blacklisted => "blocked",
            ROkStatus::RateLimit => "rate-limit",
            ROkStatus::Invalid => "invalid",
            ROkStatus::Error(_) => "error",
        }
        .to_owned()
    }

    /// Human readable status
    #[allow(clippy::inherent_to_string)]
    pub fn to_string(&self) -> String {
        let message = match self {
            ROkStatus::Ok => "",
            ROkStatus::Pow => "difficulty",
            ROkStatus::Duplicate => "already have this event",
            ROkStatus::Blocked => "try again later",
            ROkStatus::Blacklisted => "you are banned from posting here",
            ROkStatus::RateLimit => "slow down there chief",
            ROkStatus::Invalid => "event creation date is too far off from the current time",
            ROkStatus::Error(e) => e,
        }
        .to_owned();

        match self {
            ROkStatus::Ok => "".to_owned(),
            _ => {
                format!("{}:{}", self.computer_readable(), message)
            }
        }
    }
}

/// OK messages
#[derive(Clone, PartialEq, Eq, Debug)]
pub struct ROk {
    /// Event Id
    pub id: Addr,
    /// Status
    pub status: ROkStatus,
}

impl SerializeDeserialize for ROk {
    fn get_tag() -> &'static str {
        "OK"
    }

    fn serialize(&self) -> Result<Vec<Value>, String> {
        Ok(vec![
            Value::String(Self::get_tag().to_owned()),
            Value::String(self.id.to_hex()),
            Value::Bool(self.status == ROkStatus::Ok),
            Value::String(self.status.computer_readable()),
        ])
    }

    fn deserialize(args: VecDeque<Value>) -> Result<Self, String> {
        if args.len() < 3 {
            return Err("Invalid length".to_owned());
        }
        let id = args[0]
            .as_str()
            .ok_or_else(|| "Invalid type for element 1, expecting a string".to_owned())?
            .try_into()
            .map_err(|e: types::addr::Error| format!("Invalid id: {}", e))?;
        let status = args[1].as_bool().ok_or("Invalid type, expected bool")?;
        let message = args[2]
            .as_str()
            .ok_or_else(|| "Invalid message, expected string".to_owned())?
            .to_owned();
        Ok(Self {
            id,
            status: if status {
                ROkStatus::Ok
            } else {
                let part = message
                    .split(":")
                    .next()
                    .map(|x| x.trim().to_owned())
                    .unwrap_or_default();
                match part.as_str() {
                    "pow" => ROkStatus::Pow,
                    "duplicate" => ROkStatus::Duplicate,
                    "blocked" => ROkStatus::Blocked,
                    "blacklisted" => ROkStatus::Blacklisted,
                    "rate-limit" => ROkStatus::RateLimit,
                    "invalid" => ROkStatus::Invalid,
                    "error" => ROkStatus::Error(message),
                    _ => ROkStatus::Error(message),
                }
            },
        })
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::Response;

    #[test]
    fn parse() {
        let json = r#"["OK", "b1a649ebe8b435ec71d3784793f3bbf4b93e64e17568a741aecd4c7ddeafce30", true, "pow: difficulty 25>=24"]"#;
        let message: Response = serde_json::from_str(json).expect("message");
        assert!(message.as_ok().is_some());
    }

    #[test]
    fn serialize() {
        let ok_ = ROk {
            id: "a0b0".try_into().expect("valid_id"),
            status: ROkStatus::Ok,
        };
        let m: Response = ok_.into();
        assert_eq!(
            r#"["OK","a0b0",true,""]"#,
            serde_json::to_string(&m).expect("valid json")
        );
    }
}
